import numpy as np
import numba as nb
from numba.core import types
from pySimplify.simplify_wrapper import pySimplify
import trimesh as tr
from skimage import measure
from math import ceil, floor

'''
Metaball based block LOD generator.
Input data:
    Uniform 3D mesh of metaball samples, usually 8x8x8 but it is possible to use other sizes,
    such as 16x16x16, 32x32x32, etc. Each sample consists fields:
       scale - value of metaball scale 0 - 255 (0.125 .. 32.0), where 8 is x1 scale.
       color - RGBA value of metaball color 0 - 255  
    Color 3D texture size factor. Usually twice (factor=2) as metaball mesh size, e.g. 16 for 8x8x8. 
    The texture will be 16x16x16 RGBA32 texture.

Output data:
    Binary files:
    vertices_X - array of vertices coords (4 bytes floats) for X level of details (x = 0..3)
    faces_X - array of face's vertex indexes (2 bytes uint) for X level of details (x = 0..3)
    vnormals_X - array of vertexes' normal vector values (4 bytes float) for X level of details (x = 0..3)
    texture_X - flattened 3D array of 3D color texture values (4 bytes RGBA uint) for X level of details (x = 0..3)
    
'''

''' Threshold value to determine the isosurface of metaballs composition '''
THRESHOLD = 0.5
''' Width constant of metaball equation '''
WIDTH = 1.5
''' DTYPE of metaball sample with fields "scale" and "color" '''
SAMPLES = np.dtype([("scale", np.uint8), ("color", (np.uint8, 4))])


@nb.njit()
def get_points(samples, factor=2):
    ''' Get values of metaball potential function at probe points mesh with size (n+1)x(n+1)x(n+1),
        where n is a size of samples mesh multiplied by factor value.
         samples - 3d mesh of samples' scale values
         factor - factor value
    '''
    m = max(samples.shape[:3])
    n = m * factor
    s_step = 1 / m
    p_step = 1 / n
    b = WIDTH / m
    _samples = np.zeros((m, m, m), dtype=np.float32)
    _samples[:samples.shape[0], :samples.shape[1], :samples.shape[2]] = samples / 8.0
    points = np.zeros((n + 1, n + 1, n + 1), dtype=np.float32)
    for ix in range(n + 1):
        for iy in range(n + 1):
            for iz in range(n + 1):
                x = np.array([ix * p_step, iy * p_step, iz * p_step])
                for sx in range(max(0, int(floor(ix/factor-WIDTH))), min(m, int(ceil(ix/factor+WIDTH)))):
                    for sy in range(max(0, int(floor(iy/factor-WIDTH))), min(m, int(ceil(iy/factor+WIDTH)))):
                        for sz in range(max(0, int(floor(iz/factor-WIDTH))), min(m, int(ceil(iz/factor+WIDTH)))):
                            p0 = np.array([(sx + 0.5) * s_step, (sy + 0.5) * s_step, (sz + 0.5) * s_step])
                            points[ix, iy, iz] += _samples[sx, sy, sz] * np.prod(np.exp(-(2 / b * (x - p0))**2))
    return points



@nb.njit(parallel=True)
def get_colors_normals(color_samples, points):
    ''' Get colors 3D texture array and normals 3D texture array (not used now) at probe points.
         color_samples - mesh of samples' color values
         points - probe points
    '''
    m = max(color_samples.shape[:3])
    s_step = 1 / m
    b = WIDTH / m
    _color_samples = np.zeros((m, m, m, color_samples.shape[3]))
    _color_samples[:color_samples.shape[0], :color_samples.shape[1], :color_samples.shape[2]] = color_samples
    points_shape = points.shape
    _points = points.reshape(-1, 3)
    colors = np.zeros((_points.shape[0], color_samples.shape[3]), dtype=np.float32)
    normals = np.zeros((_points.shape[0], 4), dtype=np.float32)
    for i in nb.prange(len(_points)):
        x = _points[i]
        ww = 0.0
        for sx in range(max(0, int(floor(x[0]/s_step-WIDTH))), min(m, int(ceil(x[0]/s_step+WIDTH))+1)):
            for sy in range(max(0, int(floor(x[1]/s_step-WIDTH))), min(m, int(ceil(x[1]/s_step+WIDTH))+1)):
                for sz in range(max(0, int(floor(x[2]/s_step-WIDTH))), min(m, int(ceil(x[2]/s_step+WIDTH))+1)):
                    p0 = np.array([(sx + 0.5) * s_step, (sy + 0.5) * s_step, (sz + 0.5) * s_step])
                    w = np.prod(np.exp(-(2 / b * (x - p0))**2))
                    colors[i] += _color_samples[sx, sy, sz] * w
                    normals[i, :3] += (x - p0) * w
                    ww += w
        norm = np.linalg.norm(normals[i])
        colors[i] /= ww
        normals[i] /= norm if norm > 0 else 1.0
    return colors.reshape((*points_shape[:-1], color_samples.shape[3])), normals.reshape((*points_shape[:-1], 4))


def get_assets(samples, faces_count, factor=2):
    '''
        Get final mesh, color 3D texture, vertexes normals according to level of details params:
        samples - mesh of samples
        faces_counts - target faces count in final mesh
        factor - 3D color texture size factor
    '''
    m = max(samples.shape)
    points = np.zeros((2*m+3, 2*m+3, 2*m+3))
    points[1:-1, 1:-1, 1:-1] = get_points(samples["scale"])
    verts, faces, normals, _ = measure.marching_cubes(points, THRESHOLD)
    verts = verts/(2*m + 1)
    color_points = np.zeros((factor*m, factor*m, factor*m, 3), dtype=np.float32)
    color_points[..., 0], color_points[..., 1], color_points[..., 2] = np.meshgrid(np.linspace(0, 1, factor*m),
                                                                       np.linspace(0, 1, factor*m),
                                                                       np.linspace(0, 1, factor*m))
    color_array, normal_array = get_colors_normals(samples["color"], color_points)
    mesh = tr.Trimesh(vertices=verts, faces=faces, vertex_normals=-normals)
    mesh.invert()
    _simplify = pySimplify()
    _simplify.setMesh(mesh)
    _simplify.simplify_mesh(target_count=faces_count, aggressiveness=6, preserve_border=True, verbose=0)
    new_mesh = _simplify.getMesh()
    # vert_colors, vert_normals = get_colors_normals(samples["color"], new_mesh.vertices)
    return new_mesh, color_array, new_mesh.vertex_normals


def create_lod(level, samples, faces_count, factor=2):
    '''
        Creates LOD asset files: vertices, faces, normals, 3D color texture:
        level - number of LOD [0..3]
        samples - 3D mesh of metaball samples
        faces_count - target faces count
        factor - color texture size factor value
    '''
    mesh, texture3d, vert_normals = get_assets(samples, faces_count, factor=factor)
    mesh.vertices.astype(np.float32).tofile(f"vertices_{level}.bin")
    (vert_normals[:, :3]).astype(np.float32).tofile(f"vnormals_{level}.bin")
    mesh.faces.astype(np.uint16).tofile(f"faces_{level}.bin")
    texture3d.astype(np.uint8).tofile(f"texture_{level}.bin")
    # np.abs(normals3d).astype(np.float32).tofile(f"normals_{level}.bin")


def main():
    '''
        Example of box with frontal square pit
    '''
    ''' Size of metaball samples mesh '''
    m = 8
    ''' Prepare samples '''
    samples = np.zeros((m, m, m), dtype=SAMPLES)
    samples["scale"] = np.ones((m, m, m), dtype=np.float32) * 8
    samples["scale"][2:6, 2:6, 0:5] = 0
    samples["color"][..., 0], samples["color"][..., 1], samples["color"][..., 2] = np.meshgrid(np.linspace(0, 255, m),
                                                                                                np.linspace(0, 255, m),
                                                                                                np.linspace(0, 255, m))
    samples["color"][..., 3] = 255
    ''' Create LODs '''
    create_lod(0, samples, 1000, 2)
    create_lod(1, samples, 200, 2)
    create_lod(2, samples, 50, 1)


if __name__ == "__main__":
    main()

